from .base import Function
from .test_functions import three_hump_camel 
from .test_functions import rosenbrock 
from .test_functions import ackley
from .test_functions import levi
from .test_functions import himmelblau 
from .test_functions import tang
from .test_functions import holder
from .test_functions import submanifold_rosenbrock

from .base import FunctionV2
from .landscapes.sphere import Sphere
from .landscapes.thc import ThreeHumpCamel
from .landscapes.ackley import Ackley
from .landscapes.submanifold import SubmanifoldRosenbrock
from .landscapes.rosenbrock import Rosenbrock
from .landscapes.tang import StyblinskyTang
from .landscapes.levi import Levi
from .landscapes.himmelblau import Himmelblau
from .landscapes.holder import Holder

from .utils import set_all_seed
from .utils import fixed_random_seed
from .utils import fixed_numpy_seed
from .utils import fixed_torch_seed
from .utils import fixed_all_seed


__all__ = [
    'Function',
    'three_hump_camel',
    'rosenbrock',
    'ackley',
    'levi',
    'himmelblau',
    'tang',
    'holder',
    'submanifold_rosenbrock',
    
    "FunctionV2",
    "Sphere",
    "ThreeHumpCamel",
    "Ackley",
    "SubmanifoldRosenbrock",
    "Rosenbrock",
    "StyblinskyTang",
    "Levi",
    "Himmelblau",
    "Holder",

    'set_all_seed',
    'fixed_random_seed',
    'fixed_numpy_seed',
    'fixed_torch_seed',
    'fixed_all_seed',
]



from .de import DE
from .gp import GP
from .bnn import BNN
from .dgp import DGP
from .gan import WGAN_GP as GAN
from .base import Surrogate, GenSurrogate

__all__ = [
    'BNN',
    'DE',
    'DGP',
    'GAN',
    'GP',
    'Surrogate',
    'GenSurrogate',
]


import gc
import numpy as np
from .surrogate import SurrogateOptimiser


class BarycentreSurrogateOptimiser(SurrogateOptimiser):
    def __init__(self, func, surr, acqf, **kwargs):
        super(BarycentreSurrogateOptimiser, self).__init__(func, surr, acqf, **kwargs)
        self.clear_surr = kwargs['clear_surr'] if 'clear_surr' in kwargs else False
        self.surr.models_dir = f'models_{self.func.name}_{self.seed}_{"robust" if self.acqf.robust else "optimist"}'
        self.surr.checkpoints = np.arange(119, self.surr.n_epochs, 20)
    
    def get_lip(self, X, y):
        idx = np.array(np.meshgrid(np.arange(X.shape[0]), np.arange(X.shape[0]))).T.reshape(-1, 2)
        idx = idx[idx[:, 0] != idx[:, 1]]
        L = np.max(np.linalg.norm(y[idx[:, 0]] - y[idx[:, 1]], axis=-1) / np.linalg.norm(X[idx[:, 0]] - X[idx[:, 1]], axis=-1))
        return L

    def predict(self, X, y):

        self.surr.fit(X, y)
        self.acqf.L = self.get_lip(X, y)
        self.acqf.y_mu = self.surr.y_mu.item()
        self.acqf.surr = []
        for epoch in self.surr.checkpoints:
            self.acqf.surr.append(self.surr.load_model(epoch=epoch, return_model=True))
        
        next_x = self.numerical_search(x0=X[np.argmin(y)])

        if np.any(np.linalg.norm(X - next_x, axis=-1) < 1e-6):
            next_x = np.repeat(next_x.reshape(1, -1), self.num_opt_candidates, axis=0)
            next_x += np.random.normal(0, self.eps, next_x.shape)
            next_x = next_x[np.argmin(self.acqf(next_x))]
        
        if self.clear_surr:
            del self.acqf.surr
            gc.collect()
        
        return np.array([next_x])


class EnsembleBarycentreSurrogateOptimiser(SurrogateOptimiser):
    def __init__(self, func, surr, acqf, **kwargs):
        super(EnsembleBarycentreSurrogateOptimiser, self).__init__(func, surr, acqf, **kwargs)
        
        for surr in self.surr:
            surr.verbose   = self.verbose
        self.acqf.verbose   = self.verbose
    
    def get_lip(self, X, y):
        idx = np.array(np.meshgrid(np.arange(X.shape[0]), np.arange(X.shape[0]))).T.reshape(-1, 2)
        idx = idx[idx[:, 0] != idx[:, 1]]
        L = np.max(np.linalg.norm(y[idx[:, 0]] - y[idx[:, 1]], axis=-1) / np.linalg.norm(X[idx[:, 0]] - X[idx[:, 1]], axis=-1))
        return L

    def predict(self, X, y):
        
        surrs = []
        
        for surr in self.surr:
            surr.fit(X, y)
            surrs.append(surr)
        
        self.acqf.surr = surrs
        self.acqf.L = self.get_lip(X, y)
        
        x0 = None
        if self.num_opt_start == 'fmin':
            x0 = X[np.argmin(y)]
        
        next_x = self.numerical_search(x0=x0)
        
        del self.acqf.surr
        gc.collect()

        return np.array([next_x])



import os
import gc
import torch
import gpytorch
from tqdm import tqdm
from numpy import arange
from gpytorch.models.deep_gps import DeepGP, DeepGPLayer
from gpytorch.models import AbstractVariationalGP
from gpytorch.means import ConstantMean, LinearMean
from gpytorch.kernels import RBFKernel, ScaleKernel
from gpytorch.likelihoods import GaussianLikelihood
from gpytorch.distributions import MultivariateNormal
from gpytorch.mlls import VariationalELBO, DeepApproximateMLL
from gpytorch.variational import CholeskyVariationalDistribution, VariationalStrategy

from .base import Surrogate

class DGP(Surrogate):
    def __init__(self, **kwargs):
        super(DGP, self).__init__()

        self.name         = 'DGP'
        self.model        = kwargs['model'] if 'model' in kwargs else None
        self.n_epochs     = kwargs['n_epochs'] if 'n_epochs' in kwargs else 200
        self.lr           = kwargs['lr'] if 'lr' in kwargs else 1e-1
        self.verbose      = kwargs['verbose'] if 'verbose' in kwargs else 1
        self.num_inducing = kwargs['num_inducing'] if 'num_inducing' in kwargs else 64
        self.hidden_size  = kwargs['hidden_size'] if 'hidden_size' in kwargs else 128
        self.actf         = kwargs['actf'] if 'actf' in kwargs else torch.tanh
        self.means        = kwargs['means'] if 'means' in kwargs else ['linear', 'linear']
        self.scale        = kwargs['scale'] if 'scale' in kwargs else True
        self.models_dir   = kwargs['models_dir'] if 'models_dir' in kwargs else 'models'

        self.checkpoints  = kwargs['checkpoints'] if 'checkpoints' in kwargs else (self.n_epochs + 1, 1)
        if self.checkpoints[0] < 1:
            self.checkpoints[0] = int(self.n_epochs * self.checkpoints)
        self.checkpoints   = arange(self.checkpoints[0], self.n_epochs-1, self.checkpoints[1])

        self.num_outputs = 2
        
        self.gen = torch.Generator() # for reproducibility
        self.gen.manual_seed(2208060503)
    
    def make_model(self):
        return DeepGPModel(
                in_dim       = self.input_shape,
                hidden_size  = self.hidden_size,
                num_inducing = self.num_inducing,
                actf         = self.actf,
                means        = self.means,
                gen          = self.gen
            )
    
    def fit(self, X, y, epoch=None):
        
        self.input_shape = X.shape[1]
        self.model = self.make_model()
        
        X = torch.tensor(X).float()
        y = torch.tensor(y).float().squeeze()
        
        if self.scale:
            self.y_mu, self.y_std = y.mean(), y.std()
            y = (y - self.y_mu) / (self.y_std + 1e-8)
        
        self.model.train()
        self.model.likelihood.train()
        
        optimizer = torch.optim.Adam(self.model.parameters(), lr=self.lr)

        mll = DeepApproximateMLL(VariationalELBO(self.model.likelihood, self.model, num_data=y.shape[0]))

        if self.verbose > 1:
            pbar = tqdm(range(self.n_epochs), leave=False)
        else:
            pbar = range(self.n_epochs)
        
        for epoch in pbar:
            optimizer.zero_grad()
            output = self.model(X)
            loss = -mll(output, y)
            loss.mean().backward()
            optimizer.step()

            if epoch in self.checkpoints:
                self.save_model(epoch=epoch)
            
            if self.verbose > 1:
                pbar.set_description(f'Epoch {epoch + 1} - Loss: {loss.mean().item():.3f}')
    
    def predict(self, X):

        self.model.eval()
        with torch.no_grad(), gpytorch.settings.fast_pred_var():
            observed_pred = self.model.likelihood(self.model(torch.tensor(X).float()))
            mean = observed_pred.mean[0, 0, :]
            std = torch.sqrt(observed_pred.variance)[0, 0, :]
        
        if self.scale:
            mean += self.y_mu
            std *= self.y_std
        
        return mean.detach(), std.detach()
    
    def save_model(self, epoch=1):

        if not os.path.isdir(self.models_dir):
            os.mkdir(self.models_dir)
        
        torch.save(self.model.state_dict(), f'{self.models_dir}/dgp_{epoch}.pt')
    
    def load_model(self, epoch=None, return_model=False):

        model = self.make_model()

        if epoch is None:
            epoch = self.n_epochs

        model.load_state_dict(torch.load(f'{self.models_dir}/dgp_{epoch}.pt', weights_only=True))
        model.eval()

        if return_model:
            return model
        else:
            self.model = model
            del model
            gc.colect()


class SingleLayerGP(AbstractVariationalGP):
    def __init__(self, inducing_points, mean='const'):
        variational_distribution = CholeskyVariationalDistribution(inducing_points.size(0))
        variational_strategy = VariationalStrategy(self, inducing_points, variational_distribution)
        super().__init__(variational_strategy)
        self.mean_module = ConstantMean() if mean == 'const' else  LinearMean(inducing_points.size(1))
        self.covar_module = ScaleKernel(RBFKernel(ard_num_dims=inducing_points.size(1))) 

    def forward(self, x):
        mean_x = self.mean_module(x)
        covar_x = self.covar_module(x)
        return MultivariateNormal(mean_x, covar_x)


class DeepLayer(DeepGPLayer):
    def __init__(self, input_dims, output_dims, inducing_points, mean_type='constant'):
        if output_dims is None:
            batch_shape = torch.Size([])
        else:
            batch_shape = torch.Size([output_dims])

        variational_distribution = CholeskyVariationalDistribution(
            inducing_points.size(0),
            batch_shape=batch_shape
        )

        variational_strategy = VariationalStrategy(
            self,
            inducing_points,
            variational_distribution,
            learn_inducing_locations=True
        )

        super(DeepLayer, self).__init__(variational_strategy, input_dims, output_dims)

        if mean_type == 'constant':
            self.mean_module = ConstantMean(batch_shape=batch_shape)
        else:
            self.mean_module = LinearMean(input_dims)

        self.covar_module = ScaleKernel(
            RBFKernel(batch_shape=batch_shape, ard_num_dims=input_dims),
            batch_shape=batch_shape, ard_num_dims=input_dims
        )

    def forward(self, x):
        mean_x = self.mean_module(x)
        covar_x = self.covar_module(x)
        return MultivariateNormal(mean_x, covar_x)


class DeepGPModel(DeepGP):
    def __init__(self, in_dim, out_dim=None, hidden_size=16, num_inducing=22, layers=['deep', 'deep'], means=['constant', 'constant'], actf=torch.tanh, gen=None):
        super().__init__()
        
        self.actf = actf
        
        inducing_points = torch.rand(num_inducing, in_dim, generator=gen)
        output_inducing = torch.rand(num_inducing, hidden_size if layers[1]=='deep' else 1, generator=gen)
        
        self.input_layer = DeepLayer(in_dim, hidden_size, inducing_points, mean_type=means[0]) if layers[0]=='deep' else SingleLayerGP(inducing_points)
        self.output_layer =  DeepLayer(hidden_size, None, output_inducing, mean_type=means[1]) if layers[1]=='deep' else SingleLayerGP(output_inducing)
        
        self.likelihood = GaussianLikelihood()

    def forward(self, x):
        hidden_rep = self.input_layer(x).mean
        if self.actf is not None:
            hidden_rep = self.actf(hidden_rep)
        output = self.output_layer(hidden_rep)
        return output



from .utils import display
from .utils import plot_results
# from .utils import plot_func_2d


__all__ = [
    'display',
    'plot_results',
    # 'plot_func_2d',
#     'set_all_seed',
#     'fixed_random_seed',
#     'fixed_numpy_seed',
#     'fixed_torch_seed',
#     'fixed_all_seed',
]




import torch
import gpytorch
import numpy as np
from scipy.special import expit
from scipy.spatial.distance import cdist
from scipy.stats import norm, energy_distance as Wdist
from botorch.acquisition import AnalyticAcquisitionFunction
from botorch.acquisition.objective import ExpectationPosteriorTransform

from joblib import Parallel, delayed

import multiprocessing
multiprocessing.set_start_method('spawn')


class EI(AnalyticAcquisitionFunction):
    def __init__(self, model, log_transform=True):
        '''
        Expected Improvement (EI) acquisition function.
        '''
        super(EI, self).__init__(model=model, posterior_transform=ExpectationPosteriorTransform(10))
        
        self.name = 'EI'
        self.model = model
        self.log_transform = log_transform
    
    def forward(self, x, **kwargs):
        '''
        Parameters
        ----------
        x : np.array of shape (n_samples * func.n_obs, func.dim)
            Candidate points.
        
        kwargs : dict
            Keyword arguments of the surrogate model.
        
        Returns
        -------
        EI : np.array of shape (n_samples, 1)
            EI values. They are multiplied by -1.0 for optimisation perposes.
        '''
        if len(x.shape) == 1:
            x = x.reshape(1, -1)

        mu, std = self.model.predict(x, **kwargs)

        z_ = np.min(self.y) - mu
        z  = z_ / (std + 1e-8)
        z_prob, z_dens = norm.cdf(z), norm.pdf(z)

        EI = z_ * z_prob + std * z_dens

        if self.log_transform:
            return -1.0 * np.log(EI + 1e-6)
        else:
            return -1.0 * EI


class CB(AnalyticAcquisitionFunction):
    def __init__(self, model, kappa=1.0, minimise=True):# make kappa as per Abbasi-Yadkori
        '''
        Confidence Bound (CB) type acquisition function.

        Parameters
        ----------
        kappa : float, default = 2.0
            Exploration v. Exploitation coefficient.
        
        minimise : bool, default = True
            Whether the objective is minimised or not.
            If True, the acquisition function is Lower Confidence Bound (LCB).
            If False, the acquisition function is Upper Confidence Bound (UCB).
        '''
        super(CB, self).__init__(model=model)

        self.name     = 'LCB' if minimise else 'UCB'
        self.model    = model
        self.kappa    = kappa
        self.minimise = minimise
    
    def forward(self, x, **kwargs):
        '''
        Parameters
        ----------
        x : np.array of shape (n_candidates * func.n_obs, func.dim)
            Candidates points.
        
        kwargs : dict
            Keyword arguments of the surrogate model.
        
        Returns
        -------
        regret : np.array of shape (n_candidates, 1)
            CB values.
        '''
        if len(x.shape) == 1:
            x = x.reshape(1, -1)
        
        mu, std = self.model.predict(x, **kwargs)

        if self.minimise:
            regret = mu - self.kappa * std # LCB
        else:
            regret = mu + self.kappa * std # UCB
        
        return regret


class WU(AnalyticAcquisitionFunction):
    def __init__(self, model, kappa=2.0, minimise=True):
        '''
        Wasserstein Uncertainty based regret.

        Parameters
        ----------
        kappa : float, default = 2.0
            Exploration v. Exploitation coefficient.
        
        minimise : bool, default = True
            Whether the objective is minimised or not.
        '''
        super(WU, self).__init__(model=model)

        self.name     = 'WU'
        self.model    = model
        self.kappa    = kappa
        self.minimise = minimise
    
    def forward(self, x):
        '''
        Parameters
        ----------
        x : np.array of shape (n_samples * func.n_obs, func.dim)
            Candidate points.
        
        Returns
        -------
        regret : np.array of shape (n_candidates, 1)
            Wasserstein uncertainty based regret.
        '''
        if x.ndim == 1:
            x = x.reshape(1, -1)
        
        if x.shape[0] == self.y.shape[0] * self.y.shape[1]:
            y_gen = self.model.sample(x)
        else:
            y_gen = np.concatenate([self.model.sample(x) for _ in range(self.y.shape[1])])
        
        y_gen = y_gen.reshape(y_gen.shape[0]//self.y.shape[1], self.y.shape[1])

        mu = np.mean(y_gen, axis=-1)
        wu = [[Wdist(self.y[i, :], y_gen[j, :]) for i in range(self.y.shape[0])] for j in range(y_gen.shape[0])]
        wu = np.min(wu, axis=-1)

        if self.minimise:
            regret = mu - self.kappa * wu
        else:
            regret = mu + self.kappa * wu
        
        return regret


class WU_IDW(AnalyticAcquisitionFunction):
    def __init__(self, model, kappa=2.0, minimise=True, power=1.0):
        '''
        Inverse distance weighted Wasserstein Uncertainty based regret.

        Parameters
        ----------
        kappa : float, default = 2.0
            Exploration v. Exploitation coefficient.
        
        minimise : bool, default = True
            Whether the objective is minimised or not.
        '''
        super(WU_IDW, self).__init__(model=model)

        self.name     = 'WU-IDW'
        self.model    = model
        self.kappa    = kappa
        self.minimise = minimise
        self.power    = power

    def forward(self, x):
        '''
        Parameters
        ----------
        x : np.array of shape (n_samples * func.n_obs, func.dim)
            Candidate points.
        
        Returns
        -------
        regret : np.array of shape (n_candidates, 1)
            Wasserstein uncertainty based regret.
        '''
        if x.ndim == 1:
            x = x.reshape(1, -1)
        
        if x.shape[0] == self.y.shape[0] * self.y.shape[1]:
            y_gen = self.model.sample(x)
        else:
            y_gen = np.concatenate([self.model.sample(x) for _ in range(self.y.shape[1])])
        
        y_gen = y_gen.reshape(y_gen.shape[0]//self.y.shape[1], self.y.shape[1])

        mu = np.mean(y_gen, axis=-1)
        wu = [[Wdist(self.y[i, :], y_gen[j, :]) for i in range(self.y.shape[0])] for j in range(y_gen.shape[0])]
        
        dist = cdist(self.conds, x)

        weights = 1.0 / (dist + 1e-12)**self.power
        
        weights /= weights.sum(axis=0)
        
        std = np.dot(wu, weights)
        
        std = np.diag(std)

        if self.minimise:
            regret = mu - self.kappa * std
        else:
            regret = mu + self.kappa * std
        
        return regret



class PI(AnalyticAcquisitionFunction):
    def __init__(self, model, xi=0.01, minimise=True):
        """
        Probability of Improvement (PI) acquisition function.

        Parameters
        ----------
        xi : float, default=0.01
            Exploration vs exploitation trade-off parameter.
        
        minimise : bool, default=True
            Whether the objective is minimised or maximised.
        """
        super(PI, self).__init__(model=model)
        self.name = 'PI'
        self.model = model
        self.xi = xi
        self.minimise = minimise

    def forward(self, x, **kwargs):
        """
        Calculate the PI acquisition value for candidate points.

        Parameters
        ----------
        x : np.array of shape (n_candidates, func.dim)
            Candidate points.

        Returns
        -------
        PI : np.array of shape (n_candidates,)
            Probability of Improvement values.
        """
        mu, std = self.model.predict(x, **kwargs)
        best_y = np.min(self.y) if self.minimise else np.max(self.y)
        z = (best_y - mu - self.xi) / (std + 1e-8)
        pi = norm.cdf(z)
        return pi


class ES(AnalyticAcquisitionFunction):
    def __init__(self, model):
        """
        Entropy Search (ES) acquisition function.
        """
        super(ES, self).__init__(model=model)
        self.name = "EntropySearch"
        self.model = model

    def forward(self, x, **kwargs):
        """
        Compute the reduction in entropy of the posterior over the location of the global minimum.

        Parameters:
        ----------
        x : np.array of shape (n_samples, func.dim)
            Candidate points.

        Returns:
        -------
        es : np.array
            The entropy reduction at each candidate point.
        """
        mu, std = self.model.predict(x, **kwargs)
        
        p_min = norm.cdf((np.min(self.y) - mu) / (std + 1e-8))
        entropy = p_min * np.log(p_min + 1e-8) + (1 - p_min) * np.log(1 - p_min + 1e-8)
        
        return entropy


class KG(AnalyticAcquisitionFunction):
    def __init__(self, model):
        """
        Knowledge Gradient (KG) acquisition function.
        """
        super(KG, self).__init__(model=model)
        self.name = "KnowledgeGradient"
        self.model = model

    def forward(self, x, **kwargs):
        """
        Compute the Knowledge Gradient (KG) for candidate points.

        Parameters:
        ----------
        x : np.array of shape (n_samples, func.dim)
            Candidate points.

        Returns:
        -------
        kg : np.array
            KG values at the candidate points.
        """
        mu, std = self.model.predict(x, **kwargs)
        
        delta = np.min(self.y) - mu
        z = delta / (std + 1e-8)
        kg = std * (z * norm.cdf(z) + norm.pdf(z))

        return kg


class OTUCB(AnalyticAcquisitionFunction):
    def __init__(self, model, robust=False, parallel=0):
        super(OTUCB, self).__init__(model=model)
        
        self.name = 'OTUCB'
        self.model = model
        self.robust = robust
        self.verbose = 1
        self.parallel = parallel
        self.L = 1.0
        self.y_mu = 0.0
        self.w = 0.1 * np.ones(5).reshape(-1, 1)
        
    def forward(self, x):
        
        if len(x.shape) == 1:
            x = x.reshape(1, -1)
        
        mu = []
        x = torch.tensor(x).float()
        
        for model in self.model:
            with torch.no_grad(), gpytorch.settings.fast_computations():
                mu.append(model.likelihood(model(x)).mean[0, 0, :].detach().numpy())
        
        mu  = np.array(mu)
        eps = np.std(mu, axis=0)
        mu  = np.sum(self.w * (mu + self.y_mu), axis=0)
        
        if self.robust:
            return mu + self.L * eps
        else:
            return mu - self.L * eps


class GP_OTUCB(AnalyticAcquisitionFunction):
    def __init__(self, model, wf='u', ws='h', wp=0.8, robust=False, parallel=0):
        super(GP_OTUCB, self).__init__(model=model)
        
        self.name = 'GP_OTUCB'
        self.model = None
        self.robust = robust
        self.verbose = 1
        self.wf = wf
        self.parallel = parallel
        self.L = 1.0
    
    def __single_pred(self, model):
        return model.predict(self.x)[0]
        
    def forward(self, x):
        
        if len(x.shape) == 1:
            x = x.reshape(1, -1)
        
        if self.parallel:
            mu = np.array(Parallel(n_jobs=self.parallel, prefer="threads")(delayed(self.__single_pred)(model) for model in self.model))
        else:
            mu = []
            for model in self.model:
                mu.append(model.predict(x)[0])
            mu = np.array(mu)
        
        if self.wf == 'l':
            w = np.linspace(1e-2, 1, mu.shape[1])
        elif self.wf == 'c':
            w = np.sin(np.linspace(1e-2, 1.57, mu.shape[1]))**2
        elif self.wf == 's':
            w = expit(np.linspace(-3, 3, mu.shape[1]))
        elif self.wf == 'e':
            w = np.exp(np.linspace(-2, 2, mu.shape[1]))
        else:
            w = np.ones(mu.shape[1])
        
        w /= np.sum(w)
        
        mu, eps = np.sum(w * mu, axis=0), np.std(mu)
        
        if self.robust:
            return mu + self.L * eps
        else:
            return mu - self.L * eps



import torch
import random
import contextlib
import numpy as np


@contextlib.contextmanager
def fixed_random_seed(seed: int):
    """
    Context manager to set specified seed for random package

    Parameters
    ----------
    seed: int
        The seed value to set
    """
    state = random.getstate()
    random.seed(seed)
    try:
        yield
    finally:
        random.setstate(state)


@contextlib.contextmanager
def fixed_numpy_seed(seed: int):
    """
    Context manager to set specified seed for NumPy

    Parameters
    ----------
    seed: int
        The seed value to set
    """
    state = np.random.get_state()
    np.random.seed(seed)
    try:
        yield
    finally:
        np.random.set_state(state)


@contextlib.contextmanager
def fixed_torch_seed(seed: int):
    """
    Context manager to set specified for Torch
    """
    state = torch.get_rng_state()
    cuda_state = torch.cuda.get_rng_state_all()

    torch.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)

    try:
        yield
    finally:
        torch.set_rng_state(state)
        torch.cuda.set_rng_state_all(cuda_state)


def set_all_seed(seed: int):
    """
    Set specified seed for random, numpy, torch packages

    Parameters
    ----------
    seed: int
        The seed value to set
    """
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)


@contextlib.contextmanager
def fixed_all_seed(seed: int):
    """
    Context manager to set specified seed for random, numpy and torch packages

    Parameters
    ----------
    seed: int
        The seed value to set
    """
    random_state = random.getstate()
    numpy_state = np.random.get_state()
    torch_state = torch.get_rng_state()
    torch_cuda_state = torch.cuda.get_rng_state_all()

    set_all_seed(seed)

    try:
        yield
    finally:
        random.setstate(random_state)
        np.random.set_state(numpy_state)
        torch.set_rng_state(torch_state)
        torch.cuda.set_rng_state_all(torch_cuda_state)



import numpy as np
from ..base import FunctionV2


class StyblinskyTang(FunctionV2):
    """
    d-dimensional Styblinsky-Tang function
    """
    def __init__(self, dim=20, **kwargs):
        super().__init__(**kwargs)

        self.dim = dim
        self.domain = np.tile([-5, 5], reps=(dim, 1))
        self.name = f"Styblinski-Tang (dim={self.dim})"
        self.glob_min = np.full((1, dim), fill_value=-2.903534)
        self.f_min = 0.0
        self.sigma = 1.0
    
    def func(self, x: np.ndarray) -> np.ndarray:
        return np.sum(
            (x ** 4) - 16.0 * (x ** 2) + 5.0 * x,
            axis=-1,
            keepdims=True
        )  + 39.16617 * self.dim



from .base import Surrogate

import torch
import numpy as np
import torch.nn as nn
from torchensemble import AdversarialTrainingRegressor
from torch.utils.data import TensorDataset, DataLoader

# cuda = True if torch.cuda.is_available() else False
Tensor = torch.FloatTensor #torch.cuda.FloatTensor if cuda else


class Regressor(nn.Module):
    def __init__(self, n_inputs=2, n_outputs=1, hidden_size=16):
        super(Regressor, self).__init__()

        self.model = nn.Sequential(
            nn.Linear(n_inputs, hidden_size),
            nn.Tanh(),
            nn.Linear(hidden_size, n_outputs)
        )
    
    def forward(self, x):
        x = self.model(x)
        return x


class DE(Surrogate):
    def __init__(self, **kwargs):
        super(DE, self).__init__()

        self.name         = 'DE'
        self.model        = kwargs['model'] if 'model' in kwargs else None
        self.n_estimators = kwargs['n_estimators'] if 'n_estimators' in kwargs else 10
        self.hidden_size  = kwargs['hidden_size'] if 'hidden_size' in kwargs else 64
        self.n_epochs     = kwargs['n_epochs'] if 'n_epochs' in kwargs else 100
        self.lr           = kwargs['G_lr'] if 'G_lr' in kwargs else 1e-3
        self.batch_size   = kwargs['batch_size'] if 'batch_size' in kwargs else 16
        self.verbose      = kwargs['verbose'] if 'verbose' in kwargs else 1
        self.opt          = kwargs['opt'] if 'opt' in kwargs else 'Adam'
        self.weight_decay = kwargs['weight_decay'] if 'weight_decay' in kwargs else 0
        self.device       = kwargs['device'] if 'device' in kwargs else torch.device('cpu')

        # if 'scheduler' in kwargs:
        #     self.model.set_scheduler('StepLR',
        #                             step_size = kwargs['sched_step'] if 'sched_step' in kwargs else self.n_epochs//3,
        #                             gamma = kwargs['sched_gamma'] if 'sched_gamma' in kwargs else 0.1)
    
    def fit(self, X, y):

        if self.model == None:
            base_estimator = Regressor(n_inputs=X.shape[-1], n_outputs=y.shape[-1], hidden_size=self.hidden_size)
            self.model = AdversarialTrainingRegressor(estimator = base_estimator, n_estimators = self.n_estimators)
            self.model.device = self.device
        
        self.model.set_optimizer(self.opt, lr = self.lr, weight_decay = self.weight_decay)
        
        X = Tensor(X)
        y = Tensor(y)
        
        X -= torch.min(X)
        X /= torch.max(X)
        
        X_train = TensorDataset(X, y)
        
        train_loader = DataLoader(X_train, batch_size=self.batch_size, shuffle=True, drop_last=True)
        self.model.fit(train_loader, epochs=self.n_epochs, log_interval=1000, save_model=False)
    
    def predict(self, X):

        X = Tensor(X)
        X -= torch.min(X)
        X /= torch.max(X)
        
        f   = [estimator(X).detach().numpy() for estimator in self.model.estimators_]
        f   = np.concatenate(f, axis=-1)
        std = np.std(f, axis=-1)
        f   = np.mean(f, axis=-1)

        return f, std



# import jax.numpy as jnp
# from protes import protes

from ..base import Optimiser


class ProtesOptimiser(Optimiser):
    def __init__(self, func, K=10, constraint_func=None, penalty=1e6, **kwargs):
        pass
    #     """
    #     Initialize the PROTES Optimiser.

    #     Parameters:
    #     - func: The function to optimize.
    #     - K: Number of samples per iteration.
    #     - constraint_func: Function to evaluate constraints on candidates.
    #     - penalty: Penalty value for constraint violations.
    #     """
    #     super().__init__(**kwargs)
    #     self.func = func
    #     self.K = K
    #     self.constraint_func = constraint_func
    #     self.penalty = penalty
    #     self.verbose = kwargs.get("verbose", True)

    #     self.d = self.func.T
    #     self.n = 2  
    #     self.protes_model = self.initialize_protes()

    # def initialize_protes(self):
    #     if self.verbose:
    #         print(f"Initializing PROTES with d={self.d}, n={self.n}, K={self.K}")
        
    #     return lambda f: protes(f=f, d=self.d, n=self.n, m=self.K, log=self.verbose)

    # def f_batch(self, batch):
    #     func_values = jnp.array([self.func(x) for x in batch])

    #     if self.constraint_func is not None:
    #         penalties = jnp.array([self.constraint_func(x) for x in batch])
    #         func_values += penalties * self.penalty

    #     return func_values

    # def predict(self):

    #     f_batch = lambda batch: self.f_batch(jnp.array(batch))
    #     try:
    #         best_candidate, best_value = self.protes_model(f=f_batch)
    #     except Exception as e:
    #         print(f"Error during PROTES optimization: {e}")
    #         raise

    #     if self.verbose:
    #         print(f"Best candidate: {best_candidate}, Best value: {best_value}")

    #     return best_candidate, best_value



import numpy as np
from ..base import FunctionV2


class ThreeHumpCamel(FunctionV2):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.dim = 2
        self.domain = np.tile([-5, 5], reps=(2, 1))
        self.glob_min = np.zeros((1, self.dim))
        self.f_min = 0.0
        self.name = "Three-hump camel function"
    
    def func(self, x: np.ndarray) -> np.ndarray:
        xc = x[..., 0]
        yc = x[..., 1]
        res = 2.0 * (xc ** 2) - 1.05 * (xc ** 4) + (1 / 6) * (xc ** 6) + (xc * yc) + (yc ** 2)
        return np.expand_dims(res, axis=-1)
    


import gc
import time
import torch
import pickle
import numpy as np
import matplotlib.pyplot as plt
from matplotlib import cm, ticker
from botorch.optim import optimize_acqf
# from joblib import Parallel, delayed

from .base import Optimiser
from .utils import create_dir


class SurrogateOptimiser(Optimiser):
    '''
    Surrogate based black-box optimiser.

    Parameter
    ----------
    func : Callable, waggon.functions.Function
        Black-box function to be optimised.
    
    surr : waggon.surrogates.Surrogate or waggon.surrogate.GenSurrogate
        Surrogate model for the black-box function.
    
    acqf : waggon.acquisitions.Acquisition
        Acquisition function defining the optimisation strategy.
    
    max_iter : int, default = 100
        Maximum number of optimisation loop iterations.
    
    eps : float, default = 1e-1
        Epsilon-solution criterion value.
    
    error_type : {'x', 'f'}, default = 'x'
        Optimisation error type - either in argument, 'x', or function, 'f'.
    
    num_opt : bool, default = False
        Whether the acquisition function is optimised numerically or not. If False,
        the search of the next best parameters is done via direct search.
    
    eq_cons : dict, default = None
        Equality-type constraints for numerical optimisation.
    
    ineq_cons : dict, default = None
        Ineqaulity-type constraints for numerical optimisation.

    fix_candidates : bool, default = False if num_opt else True
        Whether the candidate points should be fixed or not.
    
    n_candidates_ : int, default = 1 if num_opt else 101**2
        Number of candidates points.
    
    olhs : bool, default = True
        Whether orthogonal Latin hypercube sampling (LHS) is used for choosing candidate points.
        If False, simple LHS is used.

    lhs_seed : int, default = None
        Controls the randomness of candidates sampled via (orthogonal) LHS.
    
    verbose : int, default = 1
        Controls verbosity when fitting and predicting. By default only a progress bar over the
        optimisation loop is displayed.
    
    save_results : bool, default = True
        Whether results are saved or not.
    '''
    def __init__(self, func, surr, acqf, **kwargs):
        kwargs['func'] = func
        super(SurrogateOptimiser, self).__init__(**kwargs)
        
        self.func               = func
        self.surr               = surr
        self.acqf               = acqf
        
        self.n_candidates       = kwargs['n_candidates'] if 'n_candidates' in kwargs else 10201
        self.num_opt_start      = kwargs['num_opt_start'] if 'num_opt_start' in kwargs else 'grid'
        self.num_opt_disp       = kwargs['num_opt_disp'] if 'num_opt_disp' in kwargs else False
        self.num_opt_tol        = kwargs['num_opt_tol'] if 'num_opt_tol' in kwargs else 1e-8
        self.num_opt_candidates = kwargs['num_opt_candidates'] if 'num_opt_candidates' in kwargs else 128

        self.eq_cons            = kwargs['eq_cons'] if 'eq_cons' in kwargs else None
        self.ineq_cons          = kwargs['ineq_cons'] if 'ineq_cons' in kwargs else None

        self.jitter             = kwargs['jitter'] if 'jitter' in kwargs else 1e0
        self.parallel           = kwargs['parallel'] if 'parallel' in kwargs else 0
        
        if type(self.surr) == list:
            for s in self.surr:
                s.verbose = self.verbose
        else:
            self.surr.verbose   = self.verbose

    def predict(self, X, y):
        
        self.surr.fit(X, y)
        
        self.acqf.y = y.reshape(y.shape[0]//self.func.n_obs, self.func.n_obs)
        self.acqf.conds = X[::self.func.n_obs]
        self.acqf.model = self.surr
        
        next_x, _ = optimize_acqf(self.acqf, bounds=torch.Tensor(self.func.domain),
                                          q=1, num_restarts=200, raw_samples=1000)
        
        if hasattr(self, "clear_surr") and self.clear_surr:
            del self.acqf.surr
            gc.collect()
        
        return torch.Tensor([next_x])
    
    
    def _save(self, base_dir='test_results'):
        res_path = create_dir(self.func, self.acqf.name, self.surr.name, base_dir=base_dir)

        res = {'X': self.params,
               'y': self.res,
               'err': self.errors}

        with open(f'{res_path}/{time.strftime("%d_%m_%H_%M_%S")}.pkl', 'wb') as f:
            pickle.dump(res, f)
    

    def plot_iteration_results(self, X, next_x):
        '''
        For surrogate optimiser only.
        '''
        if self.func.dim == 1:
            inter_conds = np.linspace(self.func.domain[:, 0], self.func.domain[:, 1], 121)
        else:
            inter_conds = self.create_candidates(N=10201)
        
        # transform = lambda x: np.exp(x) if self.func.log_transform else lambda x: x

        mu, _ = self.surr.predict(inter_conds)
        mu = mu
        y_true = self.func(inter_conds)
        
        if self.func.dim == 1:
            plt.plot(inter_conds, mu, label=f'Pred: {np.mean((y_true - mu)**2):.2f}', c='cornflowerblue')
            plt.plot(inter_conds, y_true, label='True', c='orange')
            # plt.scatter(X, y, c='black')
            plt.legend()
        else:
            
            y_true, mu = y_true.reshape(101, 101), mu.reshape(101, 101)
            mse = (y_true - mu)**2

            ei = self.acqf(inter_conds).reshape(101, 101)
            x_pred = inter_conds[np.argmin(ei)]
            
            def single_2d_plot(axis, f, title):
                plt.subplot(axis)
                colormap = plt.contourf(f, locator=ticker.LinearLocator(), extent=self.func.domain.flatten(), vmin=np.min(f), vmax=np.max(f))
                plt.scatter(X[:, 0], X[:, 1], color='black')
                for gb in self.func.glob_min:
                    plt.scatter(gb[0], gb[1], color='red', marker='*')
                plt.scatter(x_pred[0], x_pred[1], color='cyan', marker='*')
                plt.scatter(next_x[0], next_x[1], color='magenta', marker='*')
                plt.title(title)
                plt.colorbar(colormap)
            
            plt.figure(figsize=(24, 6))

            single_2d_plot(141, y_true, 'True Function')
            single_2d_plot(142, mu, 'Estimated Function')
            single_2d_plot(143, mse, 'MSE')
            single_2d_plot(144, ei, 'EI')
            
            plt.tight_layout()
        
        plt.show()




import os
import numpy as np


_PRIMES = np.array([2, 3, 5, 7, 11, 13, 17, 19, 23, 29, 31, 37, 41, 43, 47, 53, 59, 61, 67, 71, 73, 79, 83, 89, 97])

def _get_olhs_num(n):
    '''
    Private function to select the number of sampling points for orthogonal Latin hypercube sampling.
    '''
    return _PRIMES[_PRIMES ** 2 > n]**2

def create_dir(func, acqf_name, surr_name, base_dir='test_results'):
    
    res_path = base_dir
    if not os.path.isdir(res_path):
        os.mkdir(res_path)
    
    func_dir = func.name if func.dim == 2 else func.name + str(func.dim)

    res_path += f'/{func_dir}'
    if not os.path.isdir(res_path):
        os.mkdir(res_path)
    
    res_path += f'/{acqf_name}'
    if not os.path.isdir(res_path):
        os.mkdir(res_path)
    
    res_path += f'/{surr_name}'
    if not os.path.isdir(res_path):
        os.mkdir(res_path)
    
    return res_path


from typing import Optional
from typing import Literal

import copy
import numpy as np
import tqdm

import torch
import torch.nn as nn
import torch.optim.lr_scheduler as lr_scheduler
import torch.utils.data as data

from .base import GenSurrogate


class Generator(nn.Module):
    def __init__(self, n_inputs, n_outputs, hidden_size=64):
        super().__init__()
        
        self.model = nn.Sequential(
            nn.Linear(n_inputs, hidden_size),
            nn.Tanh(),
            nn.Linear(hidden_size, n_outputs),
        )

    def forward(self, x_cond, x_noise):
        x = torch.cat((x_cond, x_noise), dim=1)
        x = self.model(x)
        return x
    

class Discriminator(nn.Module):
    def __init__(self, n_inputs, hidden_size=64):
        super().__init__()
        
        self.model = nn.Sequential(
            nn.Linear(n_inputs, hidden_size),
            nn.Tanh(),
            nn.Linear(hidden_size, 1),
        )

    def forward(self, x):
        x = self.model(x)
        return x

    
class WGAN_GP(GenSurrogate):
    def __init__(
        self,
        G: Optional[nn.Module] = None,
        D: Optional[nn.Module] = None,
        latent_dim: int = 10,
        hidden_size: int = 64,
        batch_size: int = 8,
        n_epochs: int = 100,
        n_disc: int = 5,
        G_lr: float = 1e-4,
        D_lr: float = 1e-4,
        lambda_gp: float = 1.0,
        scheduler: bool = False,
        verbose: Literal[0, 1, 2] = 1,
        save_loss: bool = False,
        device: Literal['auto', 'cpu', 'cuda'] = 'auto'
    ):
        super().__init__()

        self.G = G
        self.D = D
        self.latent_dim = latent_dim
        self.hidden_size = hidden_size
        self.batch_size = batch_size
        self.n_epochs = n_epochs
        self.n_disc = n_disc
        self.G_lr = G_lr
        self.D_lr = D_lr
        self.lambda_gp = lambda_gp
        self.scheduler = scheduler
        self.verbose = verbose
        self.save_loss = save_loss

        self.G_loss_hist: list = []
        self.D_loss_hist: list = []
        
        self.D_scheduler = None
        self.G_scheduler = None

        if device == "auto":
            device = "cuda" if torch.cuda.is_available() else "cpu"
        self.device = torch.device(device)

        self.num_outputs = 2
    

    def gradient_pen(self, cond_data, real_data, gen_data):
        batch_size = real_data.size(0)
        alpha = torch.rand(
            batch_size, 1,
            dtype=torch.float32,
            device=self.device
        )
        inter_data = alpha * real_data + (1 - alpha) * gen_data
        inter_data.requires_grad = True

        disc_inter = self.D(torch.cat([cond_data, inter_data], dim=-1))
        fake = torch.ones(
            batch_size, 1,
            dtype=torch.float32,
            device=self.device,
            requires_grad=True
        )

        gradients, = torch.autograd.grad(
            outputs      = disc_inter,
            inputs       = inter_data,
            grad_outputs = fake,
            create_graph = True,
            retain_graph = True,
            only_inputs  = True,
        )
        
        gp = torch.mean((gradients.norm(2, dim=1) - 1).pow(2))
        return gp
        
        
    def fit(self, X, y, **kwargs):
        verbose = kwargs.pop("verbose", self.verbose)

        self.configure_models(X, y)
        self.configure_optimizers()

        x_t = torch.tensor(X, dtype=torch.float32, device=self.device)
        y_t = torch.tensor(y, dtype=torch.float32, device=self.device)
        train_dataset = data.TensorDataset(x_t, y_t)
        
        self.G.train(mode=True)
        self.D.train(mode=True)
        
        fit_loop = range(self.n_epochs)
        if self.verbose >= 2:
            fit_loop = tqdm.tqdm(fit_loop, unit="epoch", leave=True, position=2)
        
        for epoch in fit_loop:
            train_loader = data.DataLoader(
                train_dataset,
                batch_size = self.batch_size,
                shuffle=True,
            )

            for X_batch, y_batch in train_loader:
                X_batch = X_batch.to(self.device)
                y_batch = y_batch.to(self.device)

                self.train_step(X_batch, y_batch)
            
            self.on_train_epoch_end(fit_loop)
        

    def configure_models(self, x: np.ndarray, y: np.ndarray):
        if self.G is None:
            self.G = Generator(
                n_inputs = x.shape[-1] + self.latent_dim,
                n_outputs = y.shape[-1],
                hidden_size = self.hidden_size
            ).to(self.device)
        
        if self.D is None:
            self.D = Discriminator(
                n_inputs = x.shape[-1] + y.shape[-1],
                hidden_size = self.hidden_size
            ).to(self.device)

    def configure_optimizers(self):
        if not self.scheduler:
            self.G_opt = torch.optim.Adam(self.G.parameters(), lr=self.G_lr)
            self.D_opt = torch.optim.Adam(self.D.parameters(), lr=self.D_lr)
            return

        self.G_opt = torch.optim.Adam(
            self.G.parameters(), lr=self.G_lr * 100.0
        )
        self.D_opt = torch.optim.Adam(
            self.D.parameters(), lr=self.D_lr * 100.0
        )
            
        self.G_scheduler = lr_scheduler.StepLR(
            self.G_opt, step_size=self.n_epochs // 3, gamma=0.1
        )
        self.D_scheduler = lr_scheduler.StepLR(
            self.D_opt, step_size=self.n_epochs // 3, gamma=0.1
        )
    
    def on_train_epoch_end(self, fit_loop):
        if self.G_loss_hist and self.D_loss_hist and self.verbose >= 2:
            gen_loss = self.G_loss_hist[-1]
            disc_loss = self.D_loss_hist[-1]

            fit_loop.set_description(
                f"Gen. loss: {gen_loss:.3f}, Disc. loss: {disc_loss:.4f}"
            )
        
        if self.scheduler:
            self.G_scheduler.step()
            self.D_scheduler.step()
    
    def gen_step(self, x_batch: torch.Tensor, y_batch: torch.Tensor):
        batch_size = x_batch.size(0)

        noise_batch = torch.randn(
            batch_size, self.latent_dim,
            dtype=torch.float32,
            device=self.device,
            requires_grad=True
        )

        y_gen_batch = self.G(x_batch, noise_batch)
        disc_fake = self.D(torch.cat([x_batch, y_gen_batch], dim=1))
        gen_loss = -torch.mean(disc_fake)
            
        self.G_opt.zero_grad()
        gen_loss.backward()
        self.G_opt.step()
                
        if self.save_loss:
            self.G_loss_hist.append(gen_loss.item())
    
    def disc_step(self, x_batch: torch.Tensor, y_batch: torch.Tensor):
        batch_size = x_batch.size(0)

        noise_batch = torch.randn(
            batch_size, self.latent_dim,
            dtype=torch.float32,
            device=self.device,
            requires_grad=True
        )

        y_gen_batch = self.G(x_batch, noise_batch).detach()
        disc_real = self.D(torch.cat([x_batch, y_batch], dim=1))
        disc_fake = self.D(torch.cat([x_batch, y_gen_batch], dim=1))

        disc_loss = torch.mean(disc_fake) - torch.mean(disc_real)
        grad_pen = self.gradient_pen(x_batch, y_batch, y_gen_batch)
        disc_loss += self.lambda_gp * grad_pen

        self.D_opt.zero_grad()
        disc_loss.backward()
        self.D_opt.step()
                    
        if self.save_loss:
            self.D_loss_hist.append(disc_loss.item())

    def train_step(self, x_batch: torch.Tensor, y_batch: torch.Tensor):
        for _ in range(self.n_disc):
            self.disc_step(x_batch, y_batch)
        self.gen_step(x_batch, y_batch)

    def sample(self, X_cond: np.ndarray, detach=True):
        if self.G is None:
            raise ValueError("Generator is not specified!")

        batch_size = X_cond.shape[0]
        noise = torch.randn(
            batch_size, self.latent_dim,
            dtype=torch.float32,
            device=self.device,
            requires_grad=True
        )
        
        batch_cond = torch.tensor(X_cond, dtype=torch.float32, device=self.device)
        X_gen = self.G(batch_cond, noise)

        if detach:
            return X_gen.cpu().detach().numpy()
        else:
            return X_gen.cpu()

    @property
    def history(self):
        return {
            "gen_loss": copy.deepcopy(self.G_loss_hist),
            "disc_loss": copy.deepcopy(self.D_loss_hist)
        }

    @property
    def name(self):
        return "WassersteinGAN"
